/****************************************************************************
** Meta object code from reading C++ file 'mainwindow.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../kpat-26.07.80/src/mainwindow.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'mainwindow.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN10MainWindowE_t {};
} // unnamed namespace

template <> constexpr inline auto MainWindow::qt_create_metaobjectdata<qt_meta_tag_ZN10MainWindowE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "MainWindow",
        "loadGame",
        "",
        "QUrl",
        "url",
        "addToRecentFiles",
        "slotShowGameSelectionScreen",
        "slotGameSelected",
        "id",
        "newGame",
        "startRandom",
        "restart",
        "newNumberedDeal",
        "startNumbered",
        "gameId",
        "dealNumber",
        "nextDeal",
        "previousDeal",
        "saveGame",
        "showStats",
        "undoMove",
        "redoMove",
        "toggleDrop",
        "toggleHints",
        "toggleDemo",
        "toggleDemoAction",
        "active",
        "toggleMenubar",
        "toggleFullscreen",
        "fullScreen",
        "setAutoDropEnabled",
        "enabled",
        "enableSolver",
        "enable",
        "enableSounds",
        "enableRememberState",
        "slotPickRandom",
        "configureAppearance",
        "appearanceChanged",
        "helpGame",
        "updateSolverDescription",
        "text",
        "slotUpdateMoves",
        "moves",
        "slotSnapshot",
        "slotSnapshot2",
        "generateThemePreview"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'loadGame'
        QtMocHelpers::SlotData<bool(const QUrl &, bool)>(1, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 3, 4 }, { QMetaType::Bool, 5 },
        }}),
        // Slot 'slotShowGameSelectionScreen'
        QtMocHelpers::SlotData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'slotGameSelected'
        QtMocHelpers::SlotData<void(int)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'newGame'
        QtMocHelpers::SlotData<void()>(9, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'startRandom'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'loadGame'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'loadGame'
        QtMocHelpers::SlotData<void(const QUrl &)>(1, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Slot 'restart'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'newNumberedDeal'
        QtMocHelpers::SlotData<void()>(12, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'startNumbered'
        QtMocHelpers::SlotData<void(int, int)>(13, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 14 }, { QMetaType::Int, 15 },
        }}),
        // Slot 'nextDeal'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'previousDeal'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'saveGame'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'showStats'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'undoMove'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'redoMove'
        QtMocHelpers::SlotData<void()>(21, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleDrop'
        QtMocHelpers::SlotData<void()>(22, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleHints'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleDemo'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleDemoAction'
        QtMocHelpers::SlotData<void(bool)>(25, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 26 },
        }}),
        // Slot 'toggleMenubar'
        QtMocHelpers::SlotData<void()>(27, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleFullscreen'
        QtMocHelpers::SlotData<void(bool)>(28, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 29 },
        }}),
        // Slot 'setAutoDropEnabled'
        QtMocHelpers::SlotData<void(bool)>(30, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 31 },
        }}),
        // Slot 'enableSolver'
        QtMocHelpers::SlotData<void(bool)>(32, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 33 },
        }}),
        // Slot 'enableSounds'
        QtMocHelpers::SlotData<void(bool)>(34, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 33 },
        }}),
        // Slot 'enableRememberState'
        QtMocHelpers::SlotData<void(bool)>(35, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 33 },
        }}),
        // Slot 'slotPickRandom'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'configureAppearance'
        QtMocHelpers::SlotData<void()>(37, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'appearanceChanged'
        QtMocHelpers::SlotData<void()>(38, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'helpGame'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'updateSolverDescription'
        QtMocHelpers::SlotData<void(const QString &)>(40, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 41 },
        }}),
        // Slot 'slotUpdateMoves'
        QtMocHelpers::SlotData<void(int)>(42, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 43 },
        }}),
        // Slot 'slotSnapshot'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'slotSnapshot2'
        QtMocHelpers::SlotData<void()>(45, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'generateThemePreview'
        QtMocHelpers::SlotData<void()>(46, 2, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<MainWindow, qt_meta_tag_ZN10MainWindowE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject MainWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<KXmlGuiWindow::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10MainWindowE_t>.metaTypes,
    nullptr
} };

void MainWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MainWindow *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: { bool _r = _t->loadGame((*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 1: _t->slotShowGameSelectionScreen(); break;
        case 2: _t->slotGameSelected((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 3: _t->newGame(); break;
        case 4: _t->startRandom(); break;
        case 5: _t->loadGame(); break;
        case 6: _t->loadGame((*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[1]))); break;
        case 7: _t->restart(); break;
        case 8: _t->newNumberedDeal(); break;
        case 9: _t->startNumbered((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 10: _t->nextDeal(); break;
        case 11: _t->previousDeal(); break;
        case 12: _t->saveGame(); break;
        case 13: _t->showStats(); break;
        case 14: _t->undoMove(); break;
        case 15: _t->redoMove(); break;
        case 16: _t->toggleDrop(); break;
        case 17: _t->toggleHints(); break;
        case 18: _t->toggleDemo(); break;
        case 19: _t->toggleDemoAction((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 20: _t->toggleMenubar(); break;
        case 21: _t->toggleFullscreen((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 22: _t->setAutoDropEnabled((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 23: _t->enableSolver((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 24: _t->enableSounds((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 25: _t->enableRememberState((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 26: _t->slotPickRandom(); break;
        case 27: _t->configureAppearance(); break;
        case 28: _t->appearanceChanged(); break;
        case 29: _t->helpGame(); break;
        case 30: _t->updateSolverDescription((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 31: _t->slotUpdateMoves((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 32: _t->slotSnapshot(); break;
        case 33: _t->slotSnapshot2(); break;
        case 34: _t->generateThemePreview(); break;
        default: ;
        }
    }
}

const QMetaObject *MainWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MainWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.strings))
        return static_cast<void*>(this);
    return KXmlGuiWindow::qt_metacast(_clname);
}

int MainWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = KXmlGuiWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 35)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 35;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 35)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 35;
    }
    return _id;
}
QT_WARNING_POP
