/********************************************************************************
** Form generated from reading UI file 'statisticsdialog.ui'
**
** Created by: Qt User Interface Compiler version 6.12.0
**
** WARNING! All changes made in this file will be lost when recompiling UI file!
********************************************************************************/

#ifndef UI_STATISTICSDIALOG_H
#define UI_STATISTICSDIALOG_H

#include <QtCore/QVariant>
#include <QtWidgets/QApplication>
#include <QtWidgets/QGridLayout>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QLabel>
#include <QtWidgets/QSpacerItem>
#include <QtWidgets/QVBoxLayout>
#include <QtWidgets/QWidget>
#include <klocalizedstring.h>
#include "kcombobox.h"

QT_BEGIN_NAMESPACE

class Ui_GameStats
{
public:
    QVBoxLayout *verticalLayout;
    QHBoxLayout *_11;
    QLabel *textLabel1;
    KComboBox *GameType;
    QHBoxLayout *_9;
    QGridLayout *_10;
    QLabel *Won;
    QLabel *textLabel7;
    QLabel *textLabel2;
    QLabel *LoseStreak;
    QLabel *textLabel8;
    QLabel *WinStreak;
    QLabel *textLabel3;
    QLabel *Played;
    QLabel *label;
    QLabel *CurrentStreak;
    QLabel *label_2;
    QLabel *MinMoves;
    QSpacerItem *spacer;

    void setupUi(QWidget *GameStats)
    {
        if (GameStats->objectName().isEmpty())
            GameStats->setObjectName("GameStats");
        GameStats->resize(300, 267);
        verticalLayout = new QVBoxLayout(GameStats);
        verticalLayout->setObjectName("verticalLayout");
        _11 = new QHBoxLayout();
#ifndef UI_Q_OS_MACOS
        _11->setContentsMargins(0, 0, 0, 0);
#endif
        _11->setObjectName("_11");
        textLabel1 = new QLabel(GameStats);
        textLabel1->setObjectName("textLabel1");
        QSizePolicy sizePolicy(QSizePolicy::Policy::Minimum, QSizePolicy::Policy::Preferred);
        sizePolicy.setHorizontalStretch(0);
        sizePolicy.setVerticalStretch(0);
        sizePolicy.setHeightForWidth(textLabel1->sizePolicy().hasHeightForWidth());
        textLabel1->setSizePolicy(sizePolicy);

        _11->addWidget(textLabel1);

        GameType = new KComboBox(GameStats);
        GameType->setObjectName("GameType");
        QSizePolicy sizePolicy1(QSizePolicy::Policy::Expanding, QSizePolicy::Policy::Fixed);
        sizePolicy1.setHorizontalStretch(0);
        sizePolicy1.setVerticalStretch(0);
        sizePolicy1.setHeightForWidth(GameType->sizePolicy().hasHeightForWidth());
        GameType->setSizePolicy(sizePolicy1);

        _11->addWidget(GameType);


        verticalLayout->addLayout(_11);

        _9 = new QHBoxLayout();
        _9->setContentsMargins(0, 0, 0, 0);
        _9->setObjectName("_9");
        _10 = new QGridLayout();
#ifndef UI_Q_OS_MACOS
        _10->setContentsMargins(0, 0, 0, 0);
#endif
        _10->setObjectName("_10");
        Won = new QLabel(GameStats);
        Won->setObjectName("Won");

        _10->addWidget(Won, 1, 1, 1, 2);

        textLabel7 = new QLabel(GameStats);
        textLabel7->setObjectName("textLabel7");

        _10->addWidget(textLabel7, 2, 0, 1, 1);

        textLabel2 = new QLabel(GameStats);
        textLabel2->setObjectName("textLabel2");

        _10->addWidget(textLabel2, 0, 0, 1, 1);

        LoseStreak = new QLabel(GameStats);
        LoseStreak->setObjectName("LoseStreak");
        LoseStreak->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignVCenter);

        _10->addWidget(LoseStreak, 3, 1, 1, 2);

        textLabel8 = new QLabel(GameStats);
        textLabel8->setObjectName("textLabel8");

        _10->addWidget(textLabel8, 3, 0, 1, 1);

        WinStreak = new QLabel(GameStats);
        WinStreak->setObjectName("WinStreak");
        WinStreak->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignVCenter);

        _10->addWidget(WinStreak, 2, 1, 1, 2);

        textLabel3 = new QLabel(GameStats);
        textLabel3->setObjectName("textLabel3");

        _10->addWidget(textLabel3, 1, 0, 1, 1);

        Played = new QLabel(GameStats);
        Played->setObjectName("Played");
        Played->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignVCenter);

        _10->addWidget(Played, 0, 1, 1, 2);

        label = new QLabel(GameStats);
        label->setObjectName("label");

        _10->addWidget(label, 4, 0, 1, 1);

        CurrentStreak = new QLabel(GameStats);
        CurrentStreak->setObjectName("CurrentStreak");

        _10->addWidget(CurrentStreak, 4, 1, 1, 2);

        label_2 = new QLabel(GameStats);
        label_2->setObjectName("label_2");

        _10->addWidget(label_2, 5, 0, 1, 1);

        MinMoves = new QLabel(GameStats);
        MinMoves->setObjectName("MinMoves");
        MinMoves->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignVCenter);

        _10->addWidget(MinMoves, 5, 1, 1, 2);


        _9->addLayout(_10);


        verticalLayout->addLayout(_9);

        spacer = new QSpacerItem(275, 94, QSizePolicy::Policy::Minimum, QSizePolicy::Policy::Expanding);

        verticalLayout->addItem(spacer);


        retranslateUi(GameStats);

        QMetaObject::connectSlotsByName(GameStats);
    } // setupUi

    void retranslateUi(QWidget *GameStats)
    {
        GameStats->setWindowTitle(tr2i18n("Statistics", nullptr));
        textLabel1->setText(tr2i18n("Game:", nullptr));
        Won->setText(QString());
        textLabel7->setText(tr2i18n("Longest winning streak:", nullptr));
        textLabel2->setText(tr2i18n("Games played:", nullptr));
        LoseStreak->setText(QString());
        textLabel8->setText(tr2i18n("Longest losing streak:", nullptr));
        WinStreak->setText(QString());
        textLabel3->setText(tr2i18n("Games won:", nullptr));
        Played->setText(QString());
        label->setText(tr2i18n("Current streak:", nullptr));
        CurrentStreak->setText(QString());
        label_2->setText(tr2i18n("Minimal number of moves:", nullptr));
        MinMoves->setText(QString());
    } // retranslateUi

};

namespace Ui {
    class GameStats: public Ui_GameStats {};
} // namespace Ui

QT_END_NAMESPACE

#endif // STATISTICSDIALOG_H

